from matplotlib.delaunay import delaunay
import trisfm
from collections import defaultdict
import numpy as np
import util
import logging
import pmap


class TriangleGraph(object):
    """
    This is a data structure that stores a set of triangles, and maintains
    connectivity information efficiently (or at least provides an API
    so that efficiency can be added later.
    
    Notes:
    * By a triangle, we mean a triple of indices in increasing order.  (i,j,k) where i < j < k
    * Triangles are added with the add_triangles(triangles) function and the order of arrival
      is maintained in self.triangles.
    """
    
    @property
    def n_tris(self):
        """The number of triangles in this graph."""
        return len(self.triangles)
    
    def __init__(self):
        # Where the actual triangles are stored where each triangle is sorted.
        # e.g. (1,5,6) good but (5,1,6) is bad
        self.triangles = np.zeros((0, 3), int)
        
        # This is a set containing all triangles hashed as a frozenset.
        # e.g. frozenset(i,j,k) in triangles_set will tell if it is in this set.
        #
        # This allows efficient operations, instead of linear time access to the
        # array above. 
        self.triangles_set = set()
        
        # This is an index into the triangle set, so you can do
        # triangle_index[(i,j,k)] to get the index into the triangles array.
        self.triangle_index = {}

        # This can be used to find triangles adjacent to edges.
        # That is self.adjacent_tris[(point1,point2)] will
        # yield a list of (i_tri, i_point1, i_point2) tuples such that
        # self.triangles[i_tri][i_point1] == point1 and
        # self.triangles[i_tri][i_point2] == point2
        # 
        # This is useful for looking at connectivity between triangles.
        # See iter_tris_with_common_edge() for this in use.
        self.adjacent_tris = defaultdict(set)

    def add_triangles(self, tris):

        # Convert to a set of tuples.
        tris_set = set(frozenset(tri) for tri in tris)

        # Find the new ones
        new_tris_set = tris_set - self.triangles_set

        n_added = len(new_tris_set)
        # If we still have some, then add.        
        if n_added > 0:

            # Save this information so we know where to start updating
            # adjacency information.
            n_old = len(self.triangles)
            
            # Update the set.
            self.triangles_set.update(new_tris_set)
            
            # Append to end triangle array.
            tris = np.asarray(list(tri for tri in tris if frozenset(tri) in new_tris_set))
            
            # Sort each row first.
            tris.sort(axis=1)
            
            self.triangles = np.concatenate((self.triangles, tris), axis=0)
            
            # Update point adjacency.
            for i_tri in range(n_old, len(self.triangles)):

                tri = self.triangles[i_tri]
                
                self.triangle_index[frozenset(tri)] = i_tri
                
                # Update point adjacency.
                for (i_point, point) in enumerate(tri):
                    self.adjacent_tris[point].add((i_tri, i_point))
                
                # Update edge adjacency.
                for i_point1 in range(3):
                    i_point2 = (i_point1 + 1) % 3

                    # Find an edge
                    point1, point2 = tri[i_point1], tri[i_point2]
                    
                    # And update adjacency for both orientations.
                    self.adjacent_tris[(point1, point2)].add((i_tri, i_point1, i_point2))
                    self.adjacent_tris[(point2, point1)].add((i_tri, i_point2, i_point1))
                    
        return n_added
    
    def iter_tris_with_common_edge(self, tri_indices = None):
        """
        Iterates over triangles that have a common pair.
        Either yielding tuples (asarray=False) with the structure
        point1, point2, i_tri1, i_tri1_point1, i_tri1_point2, i_tri2, i_tri2_point1, i_tri2_point2
        or an array with 8 rows representing the above quantities.
        """
        
        if tri_indices == None:
            tri_indices = np.arange(len(self.triangles))
        
        adj_tris = self.adjacent_tris
        
        # Find all the ascending edges
        edges = [edge for edge in adj_tris.keys() if isinstance(edge, tuple) and len(edge) == 2 and edge[0] < edge[1]]
        
        for edge in edges:
            
            point1, point2 = edge
            
            # Find all the edge adj (i_tri, i_p1, i_p2) tuples.
            adj_tris_edge = list(adj_tris[edge])
            
            # Go through the adjacent triangles to this edge.
            for (i, (i_tri1, i_tri1_point1, i_tri1_point2)) in enumerate(adj_tris_edge):
                if i_tri1 not in tri_indices:
                    continue
                
                # And any other triangle that makes a pair i+1: to avoid duplicates and
                # self pairs.
                for (i_tri2, i_tri2_point1, i_tri2_point2) in adj_tris_edge[i + 1:]:
                    
                    if i_tri2 not in tri_indices:
                        continue
                    
                    yield (point1, point2,
                        i_tri1, i_tri1_point1, i_tri1_point2,
                        i_tri2, i_tri2_point1, i_tri2_point2)
                    
        return
    
    def get_tris_with_common_edge(self, tri_indices = None):
        return np.asarray(list(self.iter_tris_with_common_edge(tri_indices = tri_indices))).T
    
    def index_lookup(self, triangles):
        return np.asarray([self.triangle_index[frozenset(tri)] for tri in triangles])
    

class TriangleSoup(TriangleGraph):
    """A TriangleSoup is a TriangleGraph built on top of a set of tracks but that also
    maintains a set of rigid triangle models for the triangles in the graph.
    """
    
    def __init__(self, tracks, params):
        self.tracks = tracks
        self.n_frames = tracks.shape[0]
 
        self.fits = []
        
        self.params = params
        
        if params.use_memory:
            from joblib import Memory
            memory = Memory(params.soup_memory, verbose = False)
            self._trisfm_fn = memory.cache(trisfm.trisfm)
        else:
            self._trisfm_fn = trisfm.trisfm
            
        super(TriangleSoup, self).__init__()

    @property
    def num_frames(self):
        return self.n_frames
    
    @property
    def n_points(self):
        return self.tracks.shape[2]

    def get_subsoup(self, idx):
        """Get a soup, where triangles = self.triangles[idx]
        """
        
        # Create the new soup.
        soup = TriangleSoup(self.tracks, self.params)
        
        # Add the triangles, which updates all the convenience structures.
        soup.add_triangles(self.triangles[idx])
        
        # Get the subset of fit triangles.
        idx_set = set(idx)
        soup.fits = [fit for (i_fit, fit) in enumerate(self.fits) if i_fit in idx_set]

        # And update the various statistics from the fits.
        soup.update_from_fits()
        
        return soup
    
    def update_from_fits(self):
        """Takes the list of 3SFM results and provides arrays for vectorized access to all
        the corresponding properties."""
        
        n_tris = len(self.fits)
        assert n_tris == self.n_tris
        
        n_frames = self.fits[0]['fit_triangles'].shape[0]
        
        # Build the arrays.
        self.fit_triangles = np.zeros((n_frames, n_tris, 3, 3))
        self.rms = np.zeros((n_tris,))
        self.lengths = np.zeros((n_tris, 3))
        
        # Fill them in.
        for (i_tri, fit) in enumerate(self.fits):
            self.fit_triangles[:, i_tri] = fit['fit_triangles']
            self.rms[i_tri] = fit['rms']
            self.lengths[i_tri] = fit['l']
            
    def add_delaunay(self, frame= -1, subset = None):
        
        if subset == None:
            subset = np.arange(self.n_points)
        
        if frame == -1:
            for f in range(self.num_frames):
                self.add_delaunay(frame=f, subset=subset)
        else:
            # Find the subset of tracks to delaunay.
            sub_tracks = self.tracks[frame][:,subset]

            # Do delaunay
            cents, edges, tris, neighbs = delaunay(sub_tracks[0], sub_tracks[1]) #@UnusedVariable
            
            # Translate to indices into the whole track set.
            tris = subset[tris]
            
            n_added = self.add_triangles(tris)
                
            logging.debug("Added {} triangles from triangulation+dual in frame {}".format(n_added, frame))
            return tris

    def calculate(self):
        """Perform 3SFM on each triple of feature points."""

        # Local variables.        
        params = self.params
        triangles = self.triangles
        tracks = self.tracks
        
        # Progress counter.
        prog = util.Progress(len(triangles), 'fit', 'triangles')
        
        # Closure around point tracks that we will pass to each worker process.
        def f(tri):
            u = tracks[:, :, tri]
            l0 = trisfm.linear_trisfm(u)[0]
            rc = self._trisfm_fn(u, lens=l0, debug=params.verbosity, lam1=params.lam1, optimizer=params.optimizer)
            return rc

        # Map across the workers.
        for rc in pmap.pmap(f, triangles, n_procs=params.num_workers):
            self.fits += [rc]
            prog += 1

        # Update properties.
        self.update_from_fits()
