"""A module for displaying 3d Visualizations."""
from traits.api import HasTraits, Range, Instance, Int, Bool, Event, Str, List, \
    on_trait_change
from traitsui.api import View, Item, HGroup, VGroup, ButtonEditor, CheckListEditor
from tvtk.pyface.scene_editor import SceneEditor

import numpy as np
import logging
from mayavi import mlab
from pyface.gui import GUI


class Dancer(object):
    """A Dancer object currently wraps a MayaVI Actor and provides
    functionality for updating its per-frame appearence.
    
    Called because Actor is very much overloaded by MayaVI Actors as
    well as TVTK Actors.  Dancer emphasizes the ability to move.
    """
    
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        
    def display(self, mlab, frame):
        actor = self._display(mlab, frame)
        return actor
    
class ImagePlane(object):
    
    def __init__(self, z, size):
        self.z = z
        self.width, self.height = size

    def _display(self, mlab, frame):
        x, y = np.meshgrid((-1, 1), (-1, 1))
        x *= self.width / 2.0
        y *= self.height / 2.0
        z = np.ones(x.shape) * self.z
        s = mlab.mesh(x, y, z)
        return s

    def update(self, glyph, frame):
        pass

    def num_frames(self):
        return 1

class ImageSequence(Dancer):
    def __init__(self, image_collection, position='center', depth=0, *args, **kwargs):
        logging.debug('Loading Images')
        self.image_collection = image_collection
        logging.debug('%d images loaded.' % len(self.image_collection))

        # Build image cache.
        self.im_cache = [None] * len(self.image_collection)

        
        if position == 'origin':
            self.position = np.asarray([0, 0, 0])        
        elif position == 'center':
            h, w = self.image_collection[0].shape[:2]
            self.position = np.asarray([w / 2, h / 2, depth])
        else:
            self.position = position
        Dancer.__init__(self, *args, **kwargs)

    def get_im(self, f):
        im = self.im_cache[f]
        if im == None:
            im = self.image_collection[f]
            if len(im.shape)==3:
                dt = im.dtype
                im = im.mean(axis=-1).astype(dt)
            im = im.T
            self.im_cache[f] = im
        return im
    
    def _display(self, mlab, frame):
        image = mlab.imshow(self.get_im(frame))
        image.actor.position = self.position
        return image

    def update(self, glyph, frame):
        glyph.mlab_source.scalars = np.asarray(self.get_im(frame))

    def num_frames(self):
        return len(self.image_collection)

class TriangleMesh(Dancer):
    def __init__(self, points, tris=None, *args, **kwargs):
        super(TriangleMesh, self).__init__(*args, **kwargs)
        
        self.points = points
        
        if tris == None:
            from matplotlib.delaunay import delaunay
            tris = delaunay(points[0, 0, :], points[0, 1, :])[2]
        self.tris = tris

    def _display(self, mlab, frame):
        x, y, z = self.points[frame, :, :]
        return mlab.triangular_mesh(x, y, z, self.tris, **self.kwargs)

    def update(self, glyph, frame):
        x, y, z = self.points[frame, :, :]
        glyph.mlab_source.set(x=x, y=y, z=z)

    def num_frames(self):
        return self.points.shape[0]

class Triangles(TriangleMesh):

    def __init__(self, triangles, *args, **kwargs):

        if 'representation' not in kwargs:
            kwargs['representation'] = 'wireframe'
        F, n_tris = triangles.shape[0:2]
        # Construct a list of points and triangles.
        tris = np.arange(n_tris * 3).reshape(n_tris, 3)

        points = triangles.copy().swapaxes(2, 3).reshape(F, n_tris * 3, 3).swapaxes(1, 2)
        TriangleMesh.__init__(self, points, tris, *args, **kwargs)

class Points3D(Dancer):

    def __init__(self, points, size= -1, *args, **kwargs):

        # Default for size
        if size == -1:
            centre = np.mean(points[0], 0)
            dists = points[0] - centre[np.newaxis, :]
            dists = np.sqrt(dists * dists)
            size = np.mean(dists) / 25.0
        kwargs['scale_factor'] = size

        self.points = points.copy()
        Dancer.__init__(self, *args, **kwargs)

    def _display(self, mlab, frame):
        x, y, z = self.points[frame, :, :]
        return mlab.points3d(x, y, z, *self.args, **self.kwargs)

    def update(self, glyph, frame):
        x, y, z = self.points[frame, :, :]
        glyph.mlab_source.set(x=x, y=y, z=z)

    def num_frames(self):
        return self.points.shape[0]


class Visualization(HasTraits):
    from mayavi.tools.mlab_scene_model import MlabSceneModel
    scene = Instance(MlabSceneModel, ())
    is_playing = Bool(False)
    visible = List()
    dancers = List()
    selectable = List()
    selected = List()
    names = List()
    play_pause_label = Str("Play")
    play_pause_control = Event

    def __init__(self):
        HasTraits.__init__(self)
        self.elements = []
        self.dancers = []
        self._init_visible = []
        from mayavi.core.ui.mayavi_scene import MayaviScene
        self.trait_view('view', View(
        VGroup(
            Item('scene', show_label=False,
                editor=SceneEditor(scene_class=MayaviScene),
                resizable=True,
                height=600,
                width=800
            ),
            HGroup(
                Item('play_pause_control', show_label=False,
                     editor=ButtonEditor(label_value='play_pause_label')),
                Item('frame', show_label=False, springy=True)
            ),
            Item('selected', editor=CheckListEditor(values=self.selectable, cols=4), style='custom')
        ),
        buttons=['OK', 'Cancel'], resizable=True,
    ))

    def _selected_changed(self, old, new):
        for dancer in old:
            try:
                dancer.visible = False
            except:
                print 'except', dancer
        for dancer in new:
            dancer.visible = True
            element = self.elements[self.dancers.index(dancer)]
            element.update(dancer, min(self.frame, element.num_frames() - 1))
        

    def add(self, element, name=None, visible=True):
        
        self.elements += [element]
        self.names += [name]
        self._init_visible += [visible]

    @on_trait_change('frame')
    def update_plot(self):
        self.scene.disable_render = True
        for (dancer, element) in zip(self.dancers, self.elements):
            if dancer.visible:
                element.update(dancer, min(self.frame, element.num_frames() - 1))
        self.scene.disable_render = False

    def _play_pause_control_fired(self):
        print 'Playing is not implemented yet'
        if self.is_playing:
            print 'Paused'
            self.play_pause_label = 'Play'
        else:
            print 'Play'
            self.play_pause_label = 'Pause'
        self.is_playing = not self.is_playing

    def show(self, blocking=True):

        # Figure out how many frames there are maximum.
        max_frames = max(element.num_frames() for element in self.elements)

        # And add a trait for that.
        self.add_trait('frame', Range(0, max_frames - 1, mode='slider'))
        actors = []
        for (i, element) in enumerate(self.elements):
            actor = element.display(self.scene.mlab, self.frame)
            actors += [actor]
            if self.names[i] != None:
                if not self._init_visible[i]:
                    actor.visible = False
                else:
                    self.selected.append(actor)
                self.selectable.append((actor, self.names[i]))

        self.dancers = actors
        self.scene.parallel_projection = True
        
        # Construct the UI.
        self.edit_traits()
        
        # Move the camera so that x is right, y down and z is depth.
        # as to align with image coordinates.
        mlab.view(180,180)
        
        if blocking:
            # Enter event loop
            GUI().start_event_loop()
